/* 
 *	(c) 2003-2004 by Peter Wimmer
 *  pwimmer@gmx.at
 *
 *  GplMpgDec.ax is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  GplMpgDec.ax is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *  http://www.gnu.org/copyleft/gpl.html
 *
 */

#include "stdafx.h"
#include "PropertyPage.h"
#include "resource.h"

// Constructor
CMpegProperties::CMpegProperties(LPUNKNOWN pUnk, HRESULT *phr):
  CBasePropertyPage(NAME("Settings Property Page"), pUnk, IDD_SETTINGS, IDS_SETTINGSPROPERTYPAGETITLE),
  m_bIsInitialized(FALSE),
  m_pMpeg2DecFilter(NULL)
{
  DbgLog((LOG_TRACE, 1, TEXT("CMpegProperties::CMpegProperties")));
  ASSERT(phr);
}

// Provide the way for COM to create a CMpegProperties object.
CUnknown *CMpegProperties::CreateInstance(LPUNKNOWN lpunk, HRESULT *phr) {
  DbgLog((LOG_TRACE, 1, TEXT("CMpegProperties::CreateInstance")));

  CUnknown *punk = new CMpegProperties(lpunk, phr);
  if (punk == NULL) {
    *phr = E_OUTOFMEMORY;
  }
  return punk;
}

// Handles the messages for our property window.
BOOL CMpegProperties::OnReceiveMessage(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  TCHAR sz[STR_MAX_LENGTH];
  HWND hCtrl;

  switch (uMsg) {
    case WM_COMMAND:
      if (m_bIsInitialized) {
        if (LOWORD(wParam) == IDC_RESET) {
 		      SendMessage(GetDlgItem(m_Dlg, IDC_BRIGHTNESS), TBM_SETPOS, (WPARAM) 1, (LPARAM) 0);
          _stprintf(sz, "%d", SendMessage(GetDlgItem(m_Dlg, IDC_BRIGHTNESS), TBM_GETPOS, 0, 0));
          Static_SetText(GetDlgItem(m_Dlg, IDC_BRIGHTNESSLABEL), sz);
 		      SendMessage(GetDlgItem(m_Dlg, IDC_CONTRAST), TBM_SETPOS, (WPARAM) 1, (LPARAM) 100);
          _stprintf(sz, "%d", SendMessage(GetDlgItem(m_Dlg, IDC_CONTRAST), TBM_GETPOS, 0, 0));
          Static_SetText(GetDlgItem(m_Dlg, IDC_CONTRASTLABEL), sz);
 		      SendMessage(GetDlgItem(m_Dlg, IDC_HUE), TBM_SETPOS, (WPARAM) 1, (LPARAM) 0);
          _stprintf(sz, "%d", SendMessage(GetDlgItem(m_Dlg, IDC_HUE), TBM_GETPOS, 0, 0));
          Static_SetText(GetDlgItem(m_Dlg, IDC_HUELABEL), sz);
 		      SendMessage(GetDlgItem(m_Dlg, IDC_SATURATION), TBM_SETPOS, (WPARAM) 1, (LPARAM) 100);
          _stprintf(sz, "%d", SendMessage(GetDlgItem(m_Dlg, IDC_SATURATION), TBM_GETPOS, 0, 0));
          Static_SetText(GetDlgItem(m_Dlg, IDC_SATURATIONLABEL), sz);
          return (LRESULT) 0;
        } else {
          m_bDirty = TRUE;
          if (m_pPageSite) {
            m_pPageSite->OnStatusChange(PROPPAGESTATUS_DIRTY);
          }
        }
      }
      return (LRESULT) 1;
    case WM_HSCROLL:
      hCtrl = GetDlgItem(m_Dlg, IDC_BRIGHTNESS);
      if ((HWND)lParam == hCtrl) {
        _stprintf(sz, "%d", SendMessage(hCtrl, TBM_GETPOS, 0, 0));
        Static_SetText(GetDlgItem(m_Dlg, IDC_BRIGHTNESSLABEL), sz);
      }
      hCtrl = GetDlgItem(m_Dlg, IDC_CONTRAST);
      if ((HWND)lParam == hCtrl) {
        _stprintf(sz, "%d", SendMessage(hCtrl, TBM_GETPOS, 0, 0));
        Static_SetText(GetDlgItem(m_Dlg, IDC_CONTRASTLABEL), sz);
      }
      hCtrl = GetDlgItem(m_Dlg, IDC_HUE);
      if ((HWND)lParam == hCtrl) {
        _stprintf(sz, "%d", SendMessage(hCtrl, TBM_GETPOS, 0, 0));
        Static_SetText(GetDlgItem(m_Dlg, IDC_HUELABEL), sz);
      }
      hCtrl = GetDlgItem(m_Dlg, IDC_SATURATION);
      if ((HWND)lParam == hCtrl) {
        _stprintf(sz, "%d", SendMessage(hCtrl, TBM_GETPOS, 0, 0));
        Static_SetText(GetDlgItem(m_Dlg, IDC_SATURATIONLABEL), sz);
      }
      m_bDirty = TRUE;
      if (m_pPageSite) {
        m_pPageSite->OnStatusChange(PROPPAGESTATUS_DIRTY);
      }
      return (LRESULT) 0;
  }
  return CBasePropertyPage::OnReceiveMessage(hwnd,uMsg,wParam,lParam);
}

// Called when we connect to our filter.
HRESULT CMpegProperties::OnConnect(IUnknown *pUnknown) {
  DbgLog((LOG_TRACE, 1, TEXT("CMpegProperties::OnConnect")));
  ASSERT(m_pMpeg2DecFilter == NULL);

  HRESULT hr = pUnknown->QueryInterface(IID_IMpeg2DecFilter, (void **) &m_pMpeg2DecFilter);
  if (FAILED(hr))
    return E_NOINTERFACE;

  ASSERT(m_pMpeg2DecFilter);

  // Load settings. We limit the range of contrast and saturation so that
  // default point is in the middle of the slider. Values greater than 2.0
  // are useless anyway.
  m_fBrightness = m_pMpeg2DecFilter->GetBrightness();
  m_fContrast = m_pMpeg2DecFilter->GetContrast();
  if (m_fContrast > 2.0)
    m_fContrast = 2.0;
  m_fHue = m_pMpeg2DecFilter->GetHue();
  m_fSaturation = m_pMpeg2DecFilter->GetSaturation();
  if (m_fSaturation > 2.0)
    m_fSaturation = 2.0;
  m_iDeinterlaceMethod = m_pMpeg2DecFilter->GetDeinterlaceMethod();
  
  m_bIsInitialized = FALSE;

  return S_OK;
}

// Likewise called when we disconnect from our filter.
HRESULT CMpegProperties::OnDisconnect() {
  DbgLog((LOG_TRACE, 1, TEXT("CMpegProperties::OnDisconnect")));

  // Release interface.
  if (m_pMpeg2DecFilter == NULL)
    return E_UNEXPECTED;

  m_pMpeg2DecFilter->Release();
  m_pMpeg2DecFilter = NULL;

  return S_OK;
}

// We are being activated.
HRESULT CMpegProperties::OnActivate() {
  DbgLog((LOG_TRACE, 1, TEXT("CMpegProperties::OnActivate")));

  // Fill in deinterlace dropdown list.
  ComboBox_AddString(GetDlgItem(m_Dlg, IDC_DEINTERLACE), "Automatic"); 
  ComboBox_AddString(GetDlgItem(m_Dlg, IDC_DEINTERLACE), "Progessive"); 
  ComboBox_AddString(GetDlgItem(m_Dlg, IDC_DEINTERLACE), "Interlaced (Blend)"); 
  ComboBox_AddString(GetDlgItem(m_Dlg, IDC_DEINTERLACE), "Interlaced (Bob)"); 
  ComboBox_AddString(GetDlgItem(m_Dlg, IDC_DEINTERLACE), "Interlaced (Weave)"); 
  ComboBox_SetCurSel(GetDlgItem(m_Dlg, IDC_DEINTERLACE), m_iDeinterlaceMethod);

  // Set slider values.
  SendMessage(GetDlgItem(m_Dlg, IDC_BRIGHTNESS), TBM_SETRANGE, (WPARAM)1, (LPARAM)MAKELONG(-100, 100)); 
  SendMessage(GetDlgItem(m_Dlg, IDC_BRIGHTNESS), TBM_SETPOS, (WPARAM)1, (LPARAM) (m_fBrightness * 100.0 / 255.0));
  SendMessage(GetDlgItem(m_Dlg, IDC_CONTRAST), TBM_SETRANGE, (WPARAM)1, (LPARAM)MAKELONG(0, 200)); 
  SendMessage(GetDlgItem(m_Dlg, IDC_CONTRAST), TBM_SETPOS, (WPARAM)1, (LPARAM) (m_fContrast * 100.0));
  SendMessage(GetDlgItem(m_Dlg, IDC_HUE), TBM_SETRANGE, (WPARAM)1, (LPARAM)MAKELONG(-180, 180)); 
  SendMessage(GetDlgItem(m_Dlg, IDC_HUE), TBM_SETPOS, (WPARAM)1, (LPARAM)m_fHue);
  SendMessage(GetDlgItem(m_Dlg, IDC_SATURATION), TBM_SETRANGE, (WPARAM)1, (LPARAM)MAKELONG(0, 200)); 
  SendMessage(GetDlgItem(m_Dlg, IDC_SATURATION), TBM_SETPOS, (WPARAM)1, (LPARAM) (m_fSaturation * 100.0));

  TCHAR sz[STR_MAX_LENGTH];
  _stprintf(sz, "%d", (int) (m_fBrightness * 100.0 / 255));
  Static_SetText(GetDlgItem(m_Dlg, IDC_BRIGHTNESSLABEL), sz);
  _stprintf(sz, "%d", (int) (m_fContrast * 100.0));
  Static_SetText(GetDlgItem(m_Dlg, IDC_CONTRASTLABEL), sz);
  _stprintf(sz, "%d", (int) m_fHue);
  Static_SetText(GetDlgItem(m_Dlg, IDC_HUELABEL), sz);
  _stprintf(sz, "%d", (int) (m_fSaturation * 100.0));
  Static_SetText(GetDlgItem(m_Dlg, IDC_SATURATIONLABEL), sz);
  
  m_bIsInitialized = TRUE;

  return NOERROR;
}

// We are being deactivated.
HRESULT CMpegProperties::OnDeactivate(void) {
  DbgLog((LOG_TRACE, 1, TEXT("CMpegProperties::OnDeactivate")));
  m_bIsInitialized = FALSE;
  return NOERROR;
}

// Apply any changes so far made.
HRESULT CMpegProperties::OnApplyChanges()
{
  DbgLog((LOG_TRACE, 1, TEXT("CMpegProperties::OnApplyChanges")));

  GetControlValues();

  ASSERT(m_pMpeg2DecFilter);
  m_pMpeg2DecFilter->SetBrightness(m_fBrightness);
  m_pMpeg2DecFilter->SetContrast(m_fContrast);
  m_pMpeg2DecFilter->SetHue(m_fHue);
  m_pMpeg2DecFilter->SetSaturation(m_fSaturation);
  m_pMpeg2DecFilter->SetDeinterlaceMethod(m_iDeinterlaceMethod);

  return NOERROR;
}

// Read values of our controls.
void CMpegProperties::GetControlValues() {
  m_iDeinterlaceMethod = ComboBox_GetCurSel(GetDlgItem(m_Dlg, IDC_DEINTERLACE));
  m_fBrightness = SendMessage(GetDlgItem(m_Dlg, IDC_BRIGHTNESS), TBM_GETPOS, 0, 0) * 255.0 / 100.0;
  m_fContrast = SendMessage(GetDlgItem(m_Dlg, IDC_CONTRAST), TBM_GETPOS, 0, 0) / 100.0;
  m_fHue = SendMessage(GetDlgItem(m_Dlg, IDC_HUE), TBM_GETPOS, 0, 0);
  m_fSaturation = SendMessage(GetDlgItem(m_Dlg, IDC_SATURATION), TBM_GETPOS, 0, 0) / 100.0;
}
